package bdd;

import java.util.HashMap;

public class BDD 
{
	
	private HashMap<Node, Node> nodeByChildren = new HashMap<Node, Node>();
	
	public Node add(int var, Node low, Node high)
	{
		Node candidate = new Node(var, low, high);
		Node node = nodeByChildren.get(candidate);
		if (node != null) {
			return node;
		}
		nodeByChildren.put(candidate, candidate);
		return candidate;
	}
	
	public boolean contains(int var, Node low, Node high)
	{
		return nodeByChildren.containsKey(new Node(var, low, high));
	}
	
	public Node get(int var, Node low, Node high)
	{
		return nodeByChildren.get(new Node(var, low, high));
	}
	
	public Node make(int var, Node low, Node high)
	{
		if (low == high) {
			return low;
		}
		return add(var, low, high);
	}

	public Node complement(Node n)
	{
		return apply(Operators.COMPLEMENT, n, n);
	}
	
	public Node apply(Operator op, Node n1, Node n2)
	{
		HashMap<NodePair, Node> map = new HashMap<NodePair, Node>();
		return apply(op, n1, n2, map);
	}
	
	public Node apply(Operator op, Node n1, Node n2,
					  HashMap<NodePair, Node> map) 
	{
		NodePair pair = new NodePair(n1, n2);
		Node p = map.get(pair);
		if (p != null) {
			return p;
		}
		else if ((n1 == Node.ONE || n1 == Node.ZERO) 
				 && (n2 == Node.ONE || n2 == Node.ZERO)) {
			boolean v = op.op(n1 == Node.ONE ? true : false,
							  n2 == Node.ONE ? true : false);
			p = v ? Node.ONE : Node.ZERO;
		}
		else if (n1.getVar() == n2.getVar()) {
			p = make(n1.getVar(), apply(op, n1.getLow(), n2.getLow(), map),
					apply(op, n1.getHigh(), n2.getHigh(), map));
		}
		else if (n1.getVar() < n2.getVar()) {
			p = make(n1.getVar(), apply(op, n1.getLow(), n2, map),
					apply(op, n1.getHigh(), n2, map));
		}
		else {
			p = make(n1.getVar(), apply(op, n1, n2.getLow(), map),
					apply(op, n1, n2.getHigh(), map));
		}
		map.put(pair, p);
		return p;
	}
	
	public Node build(Function f)
	{
		return build(f, 0, f.getMaxVar());
	}
	
	public Node build(Function f, int var, int maxVar)
	{
		if (var > maxVar) {
			if (f.getValue()) {
				return Node.ONE;
			}
			else {
				return Node.ZERO;
			}
		}
		else {
			Node low = build(f.eval(var, false), var + 1, maxVar);
			Node high = build(f.eval(var, true), var + 1, maxVar);
			return make(var, low, high);
		}
	}
	
	public Node relProd(Node n1, Node n2)
	{
		HashMap<NodePair, Node> map = new HashMap<NodePair, Node>();
		return relProd(n1, n2, map);
	}
	
	private Node relProd(Node n1, Node n2, HashMap<NodePair, Node> map)
	{
		NodePair pair = new NodePair(n1, n2);
		Node p = map.get(pair);
		if (p != null) {
			return p;
		}
		// there is no positive valuation anymore
		else if (n1 == Node.ZERO ||  n2 == Node.ZERO) {
			return Node.ZERO;
		}
		// f is constant true and n2 != ZERO => \exists y1...yn \in G
		else if (n1 == Node.ONE) {
			return Node.ONE;
		}
		// recurse
		else if (n1.getVar() < n2.getVar()) {
			p = make(n1.getVar(), relProd(n1.getLow(), n2, map), 
					relProd(n1.getHigh(), n2, map));
		}
		// all further valuations are \in G and f != ZERO and they are at the
		// same variable
		else if (n2 == Node.ONE) {
			return Node.ONE;
		}
		else if (n1.getVar() == n2.getVar()) {
			p = apply(Operators.OR, relProd(n1.getLow(), n2.getLow(), map),
					relProd(n1.getHigh(), n2.getHigh(), map));
		}
		else {
			throw new IllegalStateException("A case that is not covered " 
											+ n1.getVar()
											+ " " + n2.getVar());
		}
		map.put(pair, p);
		return p;
	}
	
	private static class NodePair
	{
		private Node n1;
		private Node n2;
		
		public NodePair(Node n1, Node n2)
		{
			this.n1 = n1;
			this.n2 = n2;
		}

		@Override
		public boolean equals(Object obj) 
		{
			if (obj instanceof NodePair) {
				NodePair other = (NodePair)obj;
				return n1 == other.n1 && n2 == other.n2;  
			}
			return false;
		}

		@Override
		public int hashCode() 
		{
			int n1Var = n1.getVar();
			int n2Var = n2.getVar();
			return ((n1Var + n2Var) * (n1Var + n2Var + 1)) / 2 + n1Var;
		}
		
		
	}
}
